﻿using AssetBundles;
using GPUTools.Common.Scripts.Tools.Ranges;
using Leap.Unity.Attributes;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;
using UnityEngine.Rendering;
using UnityEngine.UI;
using UnityEngine.XR;

namespace AwasUI {
    internal class AwasUI : MVRScript {
        private const float LOOK_AWAY_THRESHOLD = 30f;
        private const float LOOK_UP_DOWN_THRESHOLD = 50f;
        private const int MINI_CAMERA_CULLED_LAYER = 8;
        private const int MINI_CAMERA_NOT_CULLED_LAYER = 1;

        // IMPORTANT - DO NOT make custom enums. The dynamic C# complier crashes Unity when it encounters these for
        // some reason

        private Canvas worldSpaceUiCanvas;
        private Canvas worldSpaceVrHintCanvas;
        private Canvas camAnchorCanvas;
        private bool isVrActive = XRSettings.enabled;
        private List<EventTrigger> buttonTriggers = new List<EventTrigger>();

        private AwasUIButton[] buttons;
        private AwasUIButton currentConfigButton;

        private UIDynamicButton triggerButton;
        private InputField buttonLabelInput;
        private UIDynamicColorPicker buttonColorPicker;
        private UIDynamicSlider buttonAlphaSlider;
        private UIDynamicColorPicker textColorPicker;
        private UIDynamicSlider textAlphaSlider;
        private UIDynamicToggle buttonHiddenToggle;
        private UIDynamicButton hintButton;

        private int hintButtonColorSummand = 1;
        private float prevTime = Time.time;
        private Vector2[] buttonPositions;

        private EventTrigger camPrevTrigger;
        private EventTrigger camNextTrigger;

        private Atom windowCamera = null;
        private FreeControllerV3 displayControl = null;
        private JSONStorableBool cameraOnStorable = null;
        bool initCameraOnValue = false;
        private Renderer[] windowCameraRenderers;
        private Transform[] renderersOriginalParents;
        private Shader unlitOverlayShader;
        private Shader unlitTextureShader;
        private bool isDisplayOverlay = true;
        private bool showUiInMiniCamera = false;

        private Camera mainWindowCamera;

        private float mainButtonWidth;
        private float mainButtonHeight;
        private float camButtonHeight;
        private float camButtonWidth;
        private float bannerButtonWidth;
        private int mainButtonFontSize;
        private int camButtonFontSize;

        private float mainColumn1Pos;
        private float mainColumn2Pos;
        private float camColumn3Pos;
        private float camColumn4Pos;
        private float bannerColumnPos;

        private float mainRow1Pos;
        private float mainRow2Pos;
        private float mainRow3Pos;

        private float camRow1Pos;
        private float camRow2Pos;
        private float camRow3Pos;
        private float camRow4Pos;

        private Color mainButtonColor;
        private Color mainButtonTextColor;
        private Color camButtonColor;
        private Color camButtonTextColor;

        // IMPORTANT - DO NOT OVERRIDE Awake() as it is used internally by MVRScript - instead use Init() function which
        // is called right after creation
        public override void Init() {
            try {
                // put init code in here
                SimpleTriggerHandler.LoadAssets();

                if (XRSettings.enabled) {
                    isVrActive = true;
                }

                camPrevTrigger = new EventTrigger(this, "camPrevTrigger");
                camNextTrigger = new EventTrigger(this, "camNextTrigger");

                SetButtonLayoutAndAppearance();
                CreateDefaultButtons();
                CreateWorldSpaceUi();
                CreateConfigUi();
                CreatePluginAPI();
                SetupWindowCameraAndDisplayControl();

                SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
            } catch (Exception e) {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        private void SetupWindowCameraAndDisplayControl() {
            windowCamera = SuperController.singleton.GetAtomByUid("WindowCamera");
            for (int i = 0; i < windowCamera.freeControllers.Length; i++) {
                if (windowCamera.freeControllers[i].name == "DisplayControl") {
                    displayControl = windowCamera.freeControllers[i];
                    break;
                }
            }

            JSONStorable cameraControlStorable = windowCamera.GetStorableByID("CameraControl");
            cameraOnStorable = (JSONStorableBool)cameraControlStorable.GetParam("cameraOn");
            initCameraOnValue = cameraOnStorable.val;
            cameraOnStorable.val = true;

            JSONStorable cameraScaleStorable = windowCamera.GetStorableByID("scale");
            JSONStorableFloat scale = (JSONStorableFloat)cameraScaleStorable.GetParam("scale");
            scale.val = 1.0f;

            windowCameraRenderers = windowCamera.GetComponentsInChildren<Renderer>();
            renderersOriginalParents = new Transform[windowCameraRenderers.Length];
            for (int i = 0; i < windowCameraRenderers.Length; i++) {
                renderersOriginalParents[i] = windowCameraRenderers[i].transform.parent;
                windowCameraRenderers[i].transform.SetParent(camAnchorCanvas.transform);
                windowCameraRenderers[i].gameObject.layer = MINI_CAMERA_CULLED_LAYER;
            }

            try {
                Shader[] shaders = Resources.FindObjectsOfTypeAll<Shader>();
                unlitOverlayShader = Array.Find(shaders, s => s.name == "Unlit/UnlitOverlayShader");
                unlitTextureShader = Array.Find(shaders, s => s.name == "Unlit/Texture");
                windowCameraRenderers[0].material.shader = unlitOverlayShader;
            } catch (Exception e) {
                SuperController.LogMessage(e.Message);
            }

            foreach (Camera sceneCamera in Camera.allCameras) {
                if (sceneCamera.name == "MainWindowCamera") {
                    mainWindowCamera = sceneCamera;
                }
            }
            mainWindowCamera.enabled = false;            
        }

        private void SetButtonLayoutAndAppearance() {
            float mainButtonPadding = isVrActive ? 2f : 5f;
            float camButtonPadding = isVrActive ? 1.5f : 4f;

            mainButtonWidth = isVrActive ? 45f : 90f;
            mainButtonHeight = isVrActive ? 35f : 40f;
            mainButtonFontSize = isVrActive ? 10 : 18;
            camButtonFontSize = isVrActive ? 10 : 10;
            camButtonHeight = isVrActive ? 23f : 25.3f;
            camButtonWidth = isVrActive ? 30 : 40;
            bannerButtonWidth = (mainButtonWidth * 2) + mainButtonPadding;

            mainColumn1Pos = 0;
            mainColumn2Pos = mainColumn1Pos + mainButtonWidth + mainButtonPadding;
            camColumn3Pos = mainColumn2Pos + mainButtonWidth + mainButtonPadding;
            camColumn4Pos = camColumn3Pos + camButtonWidth + camButtonPadding;
            bannerColumnPos = mainColumn1Pos;

            mainRow1Pos = 15;
            mainRow2Pos = mainRow1Pos - mainButtonHeight - mainButtonPadding;
            mainRow3Pos = mainRow2Pos - mainButtonHeight - mainButtonPadding;

            camRow1Pos = mainRow1Pos;
            camRow2Pos = camRow1Pos - camButtonHeight - camButtonPadding;
            camRow3Pos = camRow2Pos - camButtonHeight - camButtonPadding;
            camRow4Pos = mainRow3Pos;

            buttonPositions = new Vector2[4];
            buttonPositions[0] = new Vector2(mainColumn1Pos, mainRow1Pos);
            buttonPositions[1] = new Vector2(mainColumn2Pos, mainRow1Pos);
            buttonPositions[2] = new Vector2(mainColumn1Pos, mainRow2Pos);
            buttonPositions[3] = new Vector2(mainColumn2Pos, mainRow2Pos);
            
            mainButtonColor = new Color(0f, 0f, 0f, 0.25f);
            mainButtonTextColor = new Color(1f, 0.75f, 0.95f, 0.25f);
            camButtonColor = new Color(0, 0, 0, 0.25f);
            camButtonTextColor = new Color(1, 1, 1, 1);
        }

        #region API
        private void CreatePluginAPI() {
            foreach (AwasUIButton button in buttons) {
                JSONStorableString storableLabel = new JSONStorableString(button.configUiName + " button label", "", (newLabel) => {
                    button.buttonLabel = newLabel;
                }) { isRestorable = false, isStorable = false };
                RegisterString(storableLabel);

                JSONStorableColor storableButtonColor = new JSONStorableColor(button.configUiName + " button color", ColorToHSV(button.buttonColor), (h, s, v) => {
                    button.buttonColor = HSVColorPicker.HSVToRGB(new HSVColor() { H = h, S = s, V = v});
                }) { isRestorable = false, isStorable = false };
                RegisterColor(storableButtonColor);

                JSONStorableFloat storableButtonAlpha = new JSONStorableFloat(button.configUiName + " button alpha", 0.75f, (newAlpha) => {
                    button.buttonAlpha = newAlpha;
                }, 0, 1) { isRestorable = false, isStorable = false };
                RegisterFloat(storableButtonAlpha);

                JSONStorableColor storableTextColor = new JSONStorableColor(button.configUiName + " text color", ColorToHSV(button.textColor), (h, s, v) => {
                    button.textColor = HSVColorPicker.HSVToRGB(new HSVColor() { H = h, S = s, V = v });
                }) { isRestorable = false, isStorable = false };
                RegisterColor(storableTextColor);

                JSONStorableFloat storableTextAlpha = new JSONStorableFloat(button.configUiName + " text alpha", 0.75f, (newAlpha) => {
                    button.textAlpha = newAlpha;
                }, 0, 1) { isRestorable = false, isStorable = false };
                RegisterFloat(storableTextAlpha);

                JSONStorableBool storableIsHidden = new JSONStorableBool(button.configUiName + " hidden", false, (newIsHidden) => {
                    button.hidden = newIsHidden;
                }) { isRestorable = false, isStorable = false };
                RegisterBool(storableIsHidden);
            }
        }
        #endregion

        private void CreateDefaultButtons() {
            AwasUIButton topLeftButton = new AwasUIButton(this, "Top left", "TL", buttonPositions[0].x, buttonPositions[0].y);
            AwasUIButton topRightButton = new AwasUIButton(this, "Top right", "TR", buttonPositions[1].x, buttonPositions[1].y);
            AwasUIButton bottomLeftButton = new AwasUIButton(this, "Bottom left", "BL", buttonPositions[2].x, buttonPositions[2].y);
            AwasUIButton bottomRightButton = new AwasUIButton(this, "Bottom right", "BR", buttonPositions[3].x, buttonPositions[3].y);

            buttons = new AwasUIButton[] { topLeftButton, topRightButton, bottomLeftButton, bottomRightButton };
        }

        #region AwasButton
        private class AwasUIButton {
            public EventTrigger buttonTrigger;
            private string _buttonLabel;
            private Color _buttonColor;
            private float _buttonAlpha;
            private Color _textColor;
            private float _textAlpha;
            private bool _hidden = false;
            private readonly MVRScript script;
            public readonly string configUiName;
            public readonly float worldXpos;
            public readonly float worldYpos;
            private Action onUpdateCallback = () => { };

            public string buttonLabel {
                get { return _buttonLabel; }
                set {
                    _buttonLabel = value;
                    onUpdateCallback();
                }
            }

            public Color buttonColor {
                get { return _buttonColor; }
                set {
                    _buttonColor = value;
                    _buttonColor.a = _buttonAlpha;
                    onUpdateCallback();
                }
            }

            public float buttonAlpha {
                get { return _buttonAlpha; }
                set {
                    _buttonAlpha = value;
                    _buttonColor.a = value;
                    onUpdateCallback();
                }
            }

            public Color textColor {
                get { return _textColor; }
                set {
                    _textColor = value;
                    _textColor.a = _textAlpha;
                    onUpdateCallback();
                }
            }

            public float textAlpha {
                get { return _textAlpha; }
                set {
                    _textAlpha = value;
                    _textColor.a = value;
                    onUpdateCallback();
                }
            }

            public bool hidden {
                get { return _hidden; }
                set {
                    _hidden = value;
                    onUpdateCallback();
                }
            }

            public AwasUIButton(MVRScript script, string configUiName, string buttonLabel, float worldXpos, float worldYpos) {
                string triggerName = "Trigger";

                buttonTrigger = new EventTrigger(script, triggerName);
                _buttonLabel = buttonLabel;
                _buttonColor = new Color(0.23f, 0.10f, 0.18f); 
                _buttonColor.a = _buttonAlpha = 0.25f;
                _textColor = new Color(0.91f, 0.78f, 0.85f);
                _textColor.a = _textAlpha = 0.8f;
                this.script = script;
                this.configUiName = configUiName;
                this.worldXpos = worldXpos;
                this.worldYpos = worldYpos;
            }

            public void OnUpdate(Action callback) {
                onUpdateCallback = callback;
            }

            public JSONClass GetJSON() {
                JSONClass jsonClass = new JSONClass {
                    ["configUiName"] = configUiName,
                    [buttonTrigger.Name] = buttonTrigger.GetJSON(script.subScenePrefix),
                    ["buttonLabel"] = _buttonLabel,
                    ["buttonColor"] = new JSONClass {
                        ["r"] = _buttonColor.r.ToString(),
                        ["g"] = _buttonColor.g.ToString(),
                        ["b"] = _buttonColor.b.ToString(),
                    },
                    ["buttonAlpha"] = _buttonAlpha.ToString(),
                    ["textColor"] = new JSONClass {
                        ["r"] = _textColor.r.ToString(),
                        ["g"] = _textColor.g.ToString(),
                        ["b"] = _textColor.b.ToString(),
                    },
                    ["textAlpha"] = _textAlpha.ToString(),
                    ["hidden"] = _hidden.ToString()
                };

                return jsonClass;
            }

            public void RestoreFromJSON(JSONClass buttonJson, string subScenePrefix, bool isMerge, bool setMissingToDefault) {
                buttonTrigger.RestoreFromJSON(buttonJson, subScenePrefix, isMerge, setMissingToDefault);
                _buttonLabel = buttonJson["buttonLabel"];
                JSONNode jsonButtonColor = buttonJson["buttonColor"];
                _buttonColor = ParseRGBJSONFloats(jsonButtonColor);
                _buttonColor.a = _buttonAlpha = float.Parse(buttonJson["buttonAlpha"]);
                JSONNode jsonTextColor = buttonJson["textColor"];
                _textColor = ParseRGBJSONFloats(jsonTextColor);
                _textColor.a = _textAlpha = float.Parse(buttonJson["textAlpha"]);
                _hidden = bool.Parse(buttonJson["hidden"]);
                onUpdateCallback();
            }

            private Color ParseRGBJSONFloats(JSONNode jsonNode) {
                return new Color(float.Parse(jsonNode["r"]), float.Parse(jsonNode["g"]), float.Parse(jsonNode["b"]));
            }
        }
        #endregion

        #region ConfigUi
        private void CreateConfigUi() {
            List<string> buttonChoices = new List<string>();
            foreach (AwasUIButton button in buttons) {
                buttonChoices.Add(button.configUiName);
            }

            // Left
            JSONStorableStringChooser buttonSelector = new JSONStorableStringChooser("Buttons", buttonChoices, buttons[0].configUiName, "Button") {
                isRestorable = false, isStorable = false
            };
            buttonSelector.setCallbackFunction += (string choice) => SetUiForCurrentButton(GetButtonFromName(choice));
            UIDynamicPopup buttonSelectorPopup = CreateScrollablePopup(buttonSelector, false);
            buttonSelectorPopup.labelTextColor = Color.black;

            triggerButton = CreateConfigUiTriggersButton("Triggers");
            CreateConfigUiLabel("Button label");
            buttonLabelInput = CreateConfigUiTextInput();

            CreateConfigUiSpcer(10f);
            CreateConfigUiLabel("Button Color");
            buttonColorPicker = CreateConfigUiColorChooser(ColorToHSV(mainButtonColor));
            CreateConfigUiSpcer(10f);
            CreateConfigUiLabel("Button Alpha");
            buttonAlphaSlider = CreateConfigUiAlphaSlider();

            // Right
            UIDynamicButton camPrevButton = CreateButton("Cam prev trigger", true);
            UIDynamicButton camNextButton = CreateButton("Cam next trigger", true);
            camPrevButton.button.onClick.AddListener(camPrevTrigger.OpenPanelActionStart);
            camNextButton.button.onClick.AddListener(camNextTrigger.OpenPanelActionStart);

            CreateConfigUiSpcer(35f, true);
            CreateConfigUiLabel("", true);
            buttonHiddenToggle = CreateConfigUiToggle("Button hidden", true);
            CreateConfigUiSpcer(10f, true);
            CreateConfigUiLabel("Text Color", true);
            textColorPicker = CreateConfigUiColorChooser(ColorToHSV(mainButtonTextColor), true);
            CreateConfigUiSpcer(10f, true);
            CreateConfigUiLabel("Text Alpha", true);
            textAlphaSlider = CreateConfigUiAlphaSlider(true);

            SetUiForCurrentButton(buttons[0]);
        }

        private void SetUiForCurrentButton(AwasUIButton buttonToConfig) {
            currentConfigButton = buttonToConfig;

            triggerButton.button.onClick.RemoveAllListeners();
            triggerButton.button.onClick.AddListener(currentConfigButton.buttonTrigger.OpenPanelActionStart);

            buttonLabelInput.onValueChanged.RemoveAllListeners();
            buttonLabelInput.onValueChanged.AddListener(newLabel => currentConfigButton.buttonLabel = newLabel);
            buttonLabelInput.text = currentConfigButton.buttonLabel;

            buttonColorPicker.colorPicker.onColorChangedHandlers = null;
            buttonColorPicker.colorPicker.onColorChangedHandlers += (color) => currentConfigButton.buttonColor = color;
            buttonColorPicker.colorPicker.SetHSV(ColorToHSV(currentConfigButton.buttonColor));

            buttonAlphaSlider.slider.onValueChanged.RemoveAllListeners();
            buttonAlphaSlider.slider.onValueChanged.AddListener(newAlpha => currentConfigButton.buttonAlpha = newAlpha);
            buttonAlphaSlider.slider.value = currentConfigButton.buttonAlpha;

            textColorPicker.colorPicker.onColorChangedHandlers = null;
            textColorPicker.colorPicker.onColorChangedHandlers += (color) => currentConfigButton.textColor = color;
            textColorPicker.colorPicker.SetHSV(ColorToHSV(currentConfigButton.textColor));

            textAlphaSlider.slider.onValueChanged.RemoveAllListeners();
            textAlphaSlider.slider.onValueChanged.AddListener(newAlpha => currentConfigButton.textAlpha = newAlpha);
            textAlphaSlider.slider.value = currentConfigButton.textAlpha;

            buttonHiddenToggle.toggle.onValueChanged.RemoveAllListeners();
            buttonHiddenToggle.toggle.onValueChanged.AddListener(isHidden => currentConfigButton.hidden = isHidden);
            buttonHiddenToggle.toggle.isOn = currentConfigButton.hidden;
        }

        private AwasUIButton GetButtonFromName(string name) {
            AwasUIButton button = buttons[0];
            foreach (AwasUIButton awasUIButton in buttons) {
                if (awasUIButton.configUiName == name) {
                    button = awasUIButton;
                    break;
                }
            }
            return button;
        }

        private UIDynamicToggle CreateConfigUiToggle(string toggleLabel, bool rightSide) {
            return CreateToggle(new JSONStorableBool(toggleLabel, false) { isRestorable = false, isStorable = false }, rightSide);
        }

        private UIDynamicSlider CreateConfigUiAlphaSlider(bool rightSide = false) {
            UIDynamicSlider slider = CreateSlider(new JSONStorableFloat("buttonAlpha", 0.5f, 0, 1) { isStorable = false, isRestorable = false }, rightSide);
            slider.labelText.text = "Alpha";
            return slider;
        }

        private UIDynamicColorPicker CreateConfigUiColorChooser(HSVColor defaultColor, bool rightSide = false) {
            UIDynamicColorPicker colorPicker = CreateColorPicker(new JSONStorableColor("buttonColor", defaultColor) { isRestorable = false, isStorable = false }, rightSide);
            colorPicker.showLabel = false;
            return colorPicker;
        }

        private InputField CreateConfigUiTextInput() {
            JSONStorableString storableString = new JSONStorableString("", "");
            UIDynamicTextField inputTextContainer = CreateTextField(storableString);
            inputTextContainer.backgroundColor = Color.white;
            inputTextContainer.UItext.GetComponent<RectTransform>().anchoredPosition = new Vector2(10f, -10f);
            ScrollRect[] scrollRects = inputTextContainer.UItext.transform.GetComponentsInParent<ScrollRect>(true);
            for (int i = 0; i < scrollRects.Length; i++) {
                scrollRects[i].horizontal = false;
                scrollRects[i].vertical = false;
            }

            LayoutElement labelLayout = inputTextContainer.GetComponent<LayoutElement>();
            labelLayout.minHeight = labelLayout.preferredHeight = inputTextContainer.height = 50f;

            InputField input = inputTextContainer.gameObject.AddComponent<InputField>();
            input.textComponent = inputTextContainer.UItext;

            return input;
        }

        private UIDynamicButton CreateConfigUiTriggersButton(string buttonLabel) {
            return CreateButton(buttonLabel);
        }

        private void CreateConfigUiLabel(string uiLabel, bool rightSide = false) {
            JSONStorableString tempStorable = new JSONStorableString("", uiLabel);
            tempStorable.isRestorable = false;
            tempStorable.isStorable = false;

            UIDynamicTextField labelTextField = CreateTextField(tempStorable, rightSide);
            labelTextField.backgroundColor = new Color(1,1,1,0.25f);
            labelTextField.UItext.GetComponent<RectTransform>().anchoredPosition = new Vector2(10f, -10f);
            ScrollRect[] scrollRects = labelTextField.UItext.transform.GetComponentsInParent<ScrollRect>(true);
            for (int i = 0; i < scrollRects.Length; i++) {
                scrollRects[i].horizontal = false;
                scrollRects[i].vertical = false;
            }

            LayoutElement labelLayout = labelTextField.GetComponent<LayoutElement>();
            labelLayout.minHeight = labelLayout.preferredHeight = labelTextField.height = 50f;
        }

        private void CreateConfigUiSpcer(float height, bool rightSide = false) {
            UIDynamic spacer = CreateSpacer(rightSide);
            spacer.height = height;
        }

        private HSVColor ColorToHSV(Color color) {
            return HSVColorPicker.RGBToHSV(color.r, color.g, color.b);
        }
        #endregion

        #region WorldSpaceUi
        private void CreateWorldSpaceUi() {
            CreateWorldSpaceCanvases();
            foreach (AwasUIButton button in buttons) {
                UIDynamicButton worldSpaceButton = CreateWorldSpaceButtonFromAwasButton(button);
                button.OnUpdate(() => {
                    worldSpaceButton.label = button.buttonLabel;
                    if (button.hidden) {
                        worldSpaceButton.button.interactable = false;
                        worldSpaceButton.button.enabled = false;

                        worldSpaceButton.buttonColor = Color.clear;
                        worldSpaceButton.buttonText.color = Color.clear;
                    } else {
                        worldSpaceButton.button.interactable = true;
                        worldSpaceButton.button.enabled = true;

                        worldSpaceButton.buttonColor = button.buttonColor;
                        worldSpaceButton.buttonText.color = button.textColor;
                    }
                });
            }
            CreateWorldSpaceBanner();

            if (isVrActive) {
                hintButton = CreateWorldSpaceButton(worldSpaceVrHintCanvas.transform, 50f, 25f, 25f, mainRow2Pos);
                hintButton.label = "Awas";
                hintButton.buttonColor = mainButtonColor;
                hintButton.buttonText.color = mainButtonTextColor;
                hintButton.buttonText.fontSize = 12;
            }

            UIDynamicButton camPrevButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, camButtonWidth, camButtonHeight, camColumn3Pos, camRow1Pos, 12);
            UIDynamicButton camToggleButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, camButtonWidth, camButtonHeight, camColumn3Pos, camRow2Pos, camButtonFontSize);
            UIDynamicButton camNextButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, camButtonWidth, camButtonHeight, camColumn3Pos, camRow3Pos, 12);
            UIDynamicButton displayBackButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, camButtonWidth, camButtonHeight, camColumn3Pos, camRow4Pos, camButtonFontSize);
            UIDynamicButton showUiInMiniCameraButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, camButtonWidth, camButtonHeight, camColumn4Pos, camRow4Pos, camButtonFontSize);

            camPrevButton.label = "<";
            camNextButton.label = ">";
            camToggleButton.label = "Cam";
            displayBackButton.label = "Front";
            showUiInMiniCameraButton.label = "Show";

            camToggleButton.button.onClick.AddListener(ToggleMiniCameraDisplay);
            displayBackButton.button.onClick.AddListener(() => ToggleMiniCameraDisplayFrontBack(displayBackButton));
            showUiInMiniCameraButton.button.onClick.AddListener(() => ToggleUiVisibleInMiniCameraDisplay(showUiInMiniCameraButton));
            camPrevButton.button.onClick.AddListener(camPrevTrigger.Trigger);
            camNextButton.button.onClick.AddListener(camNextTrigger.Trigger);

            showUiInMiniCameraButton.buttonColor = displayBackButton.buttonColor = camPrevButton.buttonColor = camToggleButton.buttonColor = camNextButton.buttonColor = camButtonColor;
            showUiInMiniCameraButton.textColor = displayBackButton.textColor = camPrevButton.textColor = camToggleButton.textColor = camNextButton.textColor = camButtonTextColor;
        }

        private void ToggleMiniCameraDisplay() {
            cameraOnStorable.val = !cameraOnStorable.val;
            ApplyToAllMiniCamRenderers((Renderer renderer) => {
                renderer.gameObject.SetActive(!renderer.gameObject.activeSelf);
            });
        }

        private void ToggleMiniCameraDisplayFrontBack(UIDynamicButton button) {
            isDisplayOverlay = !isDisplayOverlay;
            if (isDisplayOverlay) {
                button.label = "Front";
            } else {
                button.label = "Back";
            }
            windowCameraRenderers[0].material.shader = isDisplayOverlay ? unlitOverlayShader : unlitTextureShader;
        }

        private void ToggleUiVisibleInMiniCameraDisplay(UIDynamicButton button) {
            showUiInMiniCamera = !showUiInMiniCamera;
            if (!showUiInMiniCamera) {
                button.label = "Show";
                SetWorldSpaceUiObjectsLayer(MINI_CAMERA_CULLED_LAYER);
            } else {
                button.label = "Hide";
                SetWorldSpaceUiObjectsLayer(MINI_CAMERA_NOT_CULLED_LAYER);
            }
        }

        private void SetWorldSpaceUiObjectsLayer(int layer) {
            if (worldSpaceVrHintCanvas != null) worldSpaceVrHintCanvas.gameObject.layer = layer;
            worldSpaceUiCanvas.gameObject.layer = layer;
            camAnchorCanvas.gameObject.layer = layer;
            ApplyToAllMiniCamRenderers((Renderer renderer) => {
                renderer.gameObject.layer = layer;
            });
        }

        private void ApplyToAllMiniCamRenderers(Action<Renderer> applyThisToRenderer) {
            for (int i = 0; i < windowCameraRenderers.Length; i++) {
                applyThisToRenderer(windowCameraRenderers[i]);
            }
        }

        private void CreateWorldSpaceCanvases() {
            worldSpaceUiCanvas = CreateCanvas(true);
            camAnchorCanvas = CreateCanvas(false);

            if (isVrActive) {
                worldSpaceVrHintCanvas = CreateCanvas(false);
            }
        }

        private Canvas CreateCanvas(bool withRaycaster) {
            GameObject newGameObject = new GameObject() { layer = MINI_CAMERA_CULLED_LAYER };

            float localScale = 0.001f * SuperController.singleton.worldScale;
            Canvas newCanvas = newGameObject.AddComponent<Canvas>();
            newCanvas.renderMode = RenderMode.WorldSpace;
            newCanvas.transform.localScale = new Vector3(localScale, localScale, localScale);
            SuperController.singleton.AddCanvas(newCanvas);

            CanvasScaler uiCanvasScaler = newGameObject.AddComponent<CanvasScaler>();
            if (isVrActive) {
                uiCanvasScaler.dynamicPixelsPerUnit = 4f;
            } else {
                uiCanvasScaler.dynamicPixelsPerUnit = 1f;
            }

            if (withRaycaster) {
                newGameObject.AddComponent<GraphicRaycaster>();
            }

            return newCanvas;
        }

        private UIDynamicButton CreateWorldSpaceButtonFromAwasButton(AwasUIButton button) {
            UIDynamicButton uiButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, mainButtonWidth, mainButtonHeight, button.worldXpos, button.worldYpos, mainButtonFontSize);
            uiButton.buttonColor = button.buttonColor;
            uiButton.label = button.buttonLabel;
            uiButton.buttonText.color = button.textColor;

            uiButton.button.onClick.AddListener(() => {
                button.buttonTrigger.Trigger();
            });

            return uiButton;
        }

        private UIDynamicButton CreateWorldSpaceButton(Transform parent, float width, float height, float xPos, float yPos, int fontSize = 18) {
            Transform transform = Instantiate(manager.configurableButtonPrefab.transform);
            transform.SetParent(parent, false);
            transform.gameObject.SetActive(true);
            transform.localPosition = new Vector3(xPos, yPos, 0f);

            UIDynamicButton uiButton = transform.GetComponent<UIDynamicButton>();
            uiButton.buttonText.fontSize = fontSize;

            uiButton.button.GetComponent<RectTransform>().pivot = new Vector2(0, 1);
            uiButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            uiButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);

            return uiButton;
        }

        private void CreateWorldSpaceBanner() {
            UIDynamicButton uiButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, bannerButtonWidth, camButtonHeight, bannerColumnPos, mainRow3Pos, isVrActive ? 6 : 12);
            uiButton.buttonColor = new Color(0f, 0f, 0f, 0.25f);
            uiButton.label = "https://patreon.com/awasxxx";
            uiButton.buttonText.color = new Color(1f, 0.75f, 0.95f, 0.25f);

            uiButton.button.onClick.AddListener(() => {
                SuperController.singleton.OpenLinkInBrowser(uiButton.label);
            });
        }
        #endregion

        #region Update
        private void UpdateCanvasLocalScale() {
            float localScale = 0.001f * SuperController.singleton.worldScale;
            worldSpaceUiCanvas.transform.localScale = new Vector3(localScale, localScale, localScale);
            worldSpaceVrHintCanvas.transform.localScale = new Vector3(localScale, localScale, localScale);
            camAnchorCanvas.transform.localScale = new Vector3(localScale, localScale, localScale);
        }

        private void UpdateDesktopUiPosition() {
            Transform lookCameraTransform = SuperController.singleton.lookCamera.transform;
            Vector3 lookAtPosition = lookCameraTransform.position + lookCameraTransform.forward * 100f;
            worldSpaceUiCanvas.transform.LookAt(lookAtPosition, lookCameraTransform.up);
            camAnchorCanvas.transform.position = worldSpaceUiCanvas.transform.position = Camera.main.ViewportToWorldPoint(new Vector3(0.875f, 0.2f, 1.7f));
            camAnchorCanvas.transform.rotation = worldSpaceUiCanvas.transform.rotation;

            if (windowCameraRenderers.Length > 0) {
                for (int i = 0; i < windowCameraRenderers.Length; i++) {
                    windowCameraRenderers[i].transform.position = camAnchorCanvas.transform.position + (camAnchorCanvas.transform.right * -0.22f) + (camAnchorCanvas.transform.up * -0.04f);
                    windowCameraRenderers[i].transform.rotation = worldSpaceUiCanvas.transform.rotation * Quaternion.Euler(90, 180, 0);
                }
            }
        }

        private bool isWorldSpaceUiLerping = false;
        private short prevCamRotation;
        private short currCamRotation;
        private Quaternion prevWorldSpaceUiRotation;
        private Quaternion prevWindowRendererRotation;
        private float lerpCount = 0;
        private void UpdateVrUiPositionAndVisibility() {
            prevCamRotation = currCamRotation;
            Transform worldSpaceTransform = worldSpaceUiCanvas.transform;
            Transform lookCameraTransform = SuperController.singleton.lookCamera.transform;
            Transform leftHandTransform = SuperController.singleton.leftHand.transform;
            float worldScale = SuperController.singleton.worldScale;

            Vector3 offsetPosition = (worldSpaceTransform.up * 0.15f * worldScale) + (worldSpaceTransform.right * -0.05f * worldScale) + (worldSpaceTransform.forward * -0.03f * worldScale);
            worldSpaceTransform.position = SuperController.singleton.leftHand.position + offsetPosition;

            float lookCameraUpDownAngle = Vector3.Angle(SuperController.singleton.transform.up, lookCameraTransform.up);
            float handToCameraDist2D = Vector3.Distance(new Vector3(leftHandTransform.position.x, 0, leftHandTransform.position.z), new Vector3(lookCameraTransform.position.x, 0, lookCameraTransform.position.z)) * worldScale;

            if (!isWorldSpaceUiLerping) {
                prevWorldSpaceUiRotation = worldSpaceTransform.rotation;
                prevWindowRendererRotation = windowCameraRenderers[0].transform.rotation;
            }

            bool uiIsUsingCameraRotation;
            Quaternion newWorldSpaceUiRotation;
            if (lookCameraUpDownAngle > LOOK_UP_DOWN_THRESHOLD || handToCameraDist2D <= 0.3f) {
                newWorldSpaceUiRotation = lookCameraTransform.rotation;
                uiIsUsingCameraRotation = true;
                currCamRotation = 0;
            } else {
                currCamRotation = 1;
                worldSpaceTransform.LookAt(lookCameraTransform.position);
                worldSpaceTransform.rotation *= Quaternion.Euler(0, 180, 0);
                newWorldSpaceUiRotation = worldSpaceTransform.rotation;
                worldSpaceTransform.rotation = prevWorldSpaceUiRotation;
                uiIsUsingCameraRotation = false;
            }

            if (currCamRotation != prevCamRotation) {
                lerpCount = 0;
                isWorldSpaceUiLerping = true;
            }

            TransitionToNewRotation(worldSpaceTransform, prevWorldSpaceUiRotation, newWorldSpaceUiRotation);

            camAnchorCanvas.transform.position = worldSpaceVrHintCanvas.transform.position = worldSpaceTransform.position;
            camAnchorCanvas.transform.rotation = worldSpaceVrHintCanvas.transform.rotation = worldSpaceTransform.rotation;

            Vector3 lookCameraOffset = (lookCameraTransform.forward * worldScale) + (lookCameraTransform.right * -0.1f * worldScale) + (lookCameraTransform.up * 0.3f * worldScale);
            float lookCameraAngleToUiCanvas = Vector3.Angle(worldSpaceTransform.position - lookCameraTransform.position, lookCameraOffset);
            float handToLookCameraDistance = Vector3.Distance(leftHandTransform.position, lookCameraTransform.position) * worldScale;

            if (lookCameraAngleToUiCanvas < LOOK_AWAY_THRESHOLD && handToLookCameraDistance > 0.3f) {
                worldSpaceUiCanvas.gameObject.SetActive(true);
                worldSpaceVrHintCanvas.gameObject.SetActive(false);
            } else {
                worldSpaceUiCanvas.gameObject.SetActive(false);
                worldSpaceVrHintCanvas.gameObject.SetActive(true);
            }

            Quaternion newWindowRendererRotation;            
            if (windowCameraRenderers.Length > 0) {
                Vector3 camDisplayForwardOffset = uiIsUsingCameraRotation ? Vector3.zero : (camAnchorCanvas.transform.forward * -0.05f * worldScale);
                for (int i = 0; i < windowCameraRenderers.Length; i++) {
                    windowCameraRenderers[i].transform.position = camAnchorCanvas.transform.position + (camAnchorCanvas.transform.right * -0.22f * worldScale) + (camAnchorCanvas.transform.up * -0.04f * worldScale) + camDisplayForwardOffset;
                    float camDisplayYRotation = uiIsUsingCameraRotation ? 180 : 150;
                    newWindowRendererRotation = camAnchorCanvas.transform.rotation * Quaternion.Euler(90, camDisplayYRotation, 0);
                    TransitionToNewRotation(windowCameraRenderers[i].transform, prevWindowRendererRotation, newWindowRendererRotation);
                }
            }
        }

        private void TransitionToNewRotation(Transform transform, Quaternion oldRotation, Quaternion newRotation) {
            if (isWorldSpaceUiLerping) {
                transform.rotation = Quaternion.Slerp(oldRotation, newRotation, lerpCount);
            } else {
                transform.rotation = newRotation;
            }
        }

        private float lastframe = 0;
        private void LogEvery1Sec(string message) {
            if (Time.time - lastframe > 1) {
                SuperController.LogMessage(message);
                lastframe = Time.time;
            }
        }

        private void VrHintButtonBreathingUpdate() {
            Color currentColor = hintButton.buttonColor;
            if (Time.time - prevTime >= 0.75f) {
                if (hintButtonColorSummand == 1) {
                    hintButtonColorSummand = -1;
                } else {
                    hintButtonColorSummand = 1;
                    currentColor.a = 0.05f;
                }
                prevTime = Time.time;
            }
            currentColor.a += hintButtonColorSummand * 0.01f;
            hintButton.buttonColor = currentColor;
        }

        private void UpdateUi() {
            if (isVrActive) {
                UpdateCanvasLocalScale();
                UpdateVrUiPositionAndVisibility();
            } else {
                UpdateDesktopUiPosition();
            }
        }

        void FixedUpdate() {
            try {
                lerpCount += 0.02f;
                if (lerpCount > 1f) {
                    isWorldSpaceUiLerping = false;
                    lerpCount = 1f;
                }
                if (isVrActive) {
                    VrHintButtonBreathingUpdate();
                }
            } catch (Exception e) {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        // Update is called with each rendered frame by Unity
        void Update() {
            try {
                UpdateUi();
            } catch (Exception e) {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        #endregion

        // OnDestroy is where you should put any cleanup
        // if you registered objects to supercontroller or atom, you should unregister them here
        void OnDestroy() {
            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
            mainWindowCamera.enabled = true;

            if (windowCameraRenderers.Length > 0) {
                for (int i = 0; i < windowCameraRenderers.Length; i++) {
                    windowCameraRenderers[i].transform.SetParent(renderersOriginalParents[i]);
                    windowCameraRenderers[i].gameObject.SetActive(true);
                    windowCameraRenderers[i].transform.position = displayControl.transform.position;
                    windowCameraRenderers[i].transform.rotation = displayControl.transform.rotation;
                    windowCameraRenderers[i].gameObject.layer = MINI_CAMERA_NOT_CULLED_LAYER;
                }
                windowCameraRenderers[0].transform.rotation *= Quaternion.Euler(180, 0, 0);
                windowCameraRenderers[1].transform.rotation *= Quaternion.Euler(0, 180, 0);

                windowCameraRenderers[0].material.shader = unlitTextureShader;
            }
            cameraOnStorable.val = initCameraOnValue;

            camPrevTrigger.Remove();
            camNextTrigger.Remove();
            foreach (EventTrigger eventTrigger in buttonTriggers) {
                eventTrigger.Remove();
            }
            if (worldSpaceUiCanvas != null) {
                SuperController.singleton.RemoveCanvas(worldSpaceUiCanvas);
                Destroy(worldSpaceUiCanvas.gameObject);
            }
            if (worldSpaceVrHintCanvas != null) {
                SuperController.singleton.RemoveCanvas(worldSpaceVrHintCanvas);
                Destroy(worldSpaceVrHintCanvas.gameObject);
            }
        }

        #region Boring
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false) {
            JSONClass jsonClass = base.GetJSON(includePhysical, includeAppearance, forceStore);
            if (includePhysical || forceStore) {
                needsStore = true;

                JSONArray buttonsList = new JSONArray();
                foreach (AwasUIButton button in buttons) {
                    buttonsList.Add("", button.GetJSON());
                }
                jsonClass["buttons"] = buttonsList;
                jsonClass["camPrevTrigger"] = camPrevTrigger.GetJSON();
                jsonClass["camNextTrigger"] = camNextTrigger.GetJSON();
            }

            return jsonClass;
        }

        public override void LateRestoreFromJSON(JSONClass jsonClass, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true) {
            base.LateRestoreFromJSON(jsonClass, restorePhysical, restoreAppearance, setMissingToDefault);
            if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger")) {
                foreach (AwasUIButton button in buttons) {
                    button.buttonTrigger.Remove();
                }

                JSONArray buttonsList = jsonClass["buttons"].AsArray;
                for (int i = 0; i < buttonsList.Count; ++i) {
                    JSONClass buttonJson = buttonsList[i].AsObject;
                    AwasUIButton button = buttons[i];
                    button.RestoreFromJSON(buttonJson, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
                }
                camPrevTrigger.RestoreFromJSON(jsonClass, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
                camNextTrigger.RestoreFromJSON(jsonClass, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
            }
            SetUiForCurrentButton(buttons[0]);
        }

        private void OnAtomRename(string oldid, string newid) {
            foreach (AwasUIButton button in buttons) {
                button.buttonTrigger.SyncAtomNames();
            }
        }
    }

    public class SimpleTriggerHandler : TriggerHandler {
        public static bool Loaded { get; private set; }

        private static SimpleTriggerHandler myInstance;

        private RectTransform myTriggerActionsPrefab;
        private RectTransform myTriggerActionMiniPrefab;
        private RectTransform myTriggerActionDiscretePrefab;
        private RectTransform myTriggerActionTransitionPrefab;

        public static SimpleTriggerHandler Instance {
            get {
                if (myInstance == null) {
                    myInstance = new SimpleTriggerHandler();
                }
                return myInstance;
            }
        }

        public static void LoadAssets() {
            SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
        }

        private IEnumerator LoadAssetsInternal() {
            foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", prefab => myTriggerActionsPrefab = prefab)) yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", prefab => myTriggerActionMiniPrefab = prefab)) yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", prefab => myTriggerActionDiscretePrefab = prefab)) yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", prefab => myTriggerActionTransitionPrefab = prefab)) yield return x;
            Loaded = true;
        }

        private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign) {
            AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
            if (request == null) {
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
            }
            yield return request;
            GameObject go = request.GetAsset<GameObject>();
            if (go == null) {
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
            }
            RectTransform prefab = go.GetComponent<RectTransform>();
            if (prefab == null) {
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
            }
            assign(prefab);
        }


        void TriggerHandler.RemoveTrigger(Trigger t) {
            // nothing to do
        }

        void TriggerHandler.DuplicateTrigger(Trigger t) {
            throw new NotImplementedException();
        }

        RectTransform TriggerHandler.CreateTriggerActionsUI() {
            return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionMiniUI() {
            return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionDiscreteUI() {
            return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionTransitionUI() {
            return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
        }

        void TriggerHandler.RemoveTriggerActionUI(RectTransform rt) {
            UnityEngine.Object.Destroy(rt?.gameObject);
        }
    }

    // Wrapper for easier handling of custom triggers.
    public class EventTrigger : Trigger {
        public string Name {
            get { return name; }
            set { name = value; myNeedInit = true; }
        }

        public MVRScript Owner {
            get; private set;
        }

        private string name;
        private bool myNeedInit = true;

        public EventTrigger(MVRScript owner, string name) {
            Name = name;
            Owner = owner;
            handler = SimpleTriggerHandler.Instance;
        }

        public void OpenPanelActionStart() {
            if (!SimpleTriggerHandler.Loaded) {
                SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
                return;
            }

            triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();
            if (myNeedInit) {
                triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
                triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
                triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
                triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
                myNeedInit = false;
            }
        }

        public void RestoreFromJSON(JSONClass jsonClass, string subScenePrefix, bool isMerge, bool setMissingToDefault) {
            if (jsonClass.HasKey(Name)) {
                JSONClass triggerJson = jsonClass[Name].AsObject;
                if (triggerJson != null) {
                    base.RestoreFromJSON(triggerJson, subScenePrefix, isMerge);
                }
            } else if (setMissingToDefault) {
                base.RestoreFromJSON(new JSONClass());
            }
        }

        public void Trigger() {
            active = true;
            active = false;
        }

        public void TriggerTransition(float val) {
            this.transitionInterpValue = val;
        }
    }
    #endregion
}
